﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.DebugEngine {

  [StorableClass]
  public class OperatorTrace : ObservableList<IOperator>, IContent, IDeepCloneable {

    #region fields

    [Storable]
    protected Dictionary<IAtomicOperation, IAtomicOperation> parents;
    #endregion

    #region Constructors & Cloning

    public OperatorTrace() {
        parents = new Dictionary<IAtomicOperation, IAtomicOperation>();
    }

    public OperatorTrace(int capacity) : base(capacity) {
        parents = new Dictionary<IAtomicOperation, IAtomicOperation>();
    }

    public OperatorTrace(IEnumerable<IOperator> collection): base(collection) {
        parents = new Dictionary<IAtomicOperation, IAtomicOperation>();
    }

    [StorableConstructor]
    protected OperatorTrace(bool deserializing) : base(deserializing) { }

    protected OperatorTrace(OperatorTrace original, Cloner cloner) {
      cloner.RegisterClonedObject(original, this);
      AddRange(original.Select(op => cloner.Clone(op)));
      parents = original.parents.ToDictionary(kvp => cloner.Clone(kvp.Key), kvp => cloner.Clone(kvp.Value));
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public virtual IDeepCloneable Clone(Cloner cloner) {
      return new OperatorTrace(this, cloner);
    }
    #endregion

    #region Additional List Modifiers

    public virtual void ReplaceAll(IEnumerable<IOperator> operators) {
      var oldList = list;
      list = new List<IOperator>(operators);
      if (oldList.Count != list.Count)
        OnPropertyChanged("Count");
      OnPropertyChanged("Item[]");
      OnCollectionReset(
        list.Select((op, i) => new IndexedItem<IOperator>(i, op)),
        oldList.Select((op, i) => new IndexedItem<IOperator>(i, op)));
    }

    #endregion

    #region Parent Tracing

    public virtual void RegisterParenthood(IAtomicOperation parent, IOperation children) {
      OperationCollection operations = children as OperationCollection;
      if (operations != null)
        foreach (var op in operations)
          RegisterParenthood(parent, op);
      IAtomicOperation atomicOperation = children as IAtomicOperation;
      if (atomicOperation != null && atomicOperation.Operator != null && !parents.ContainsKey(atomicOperation))
        parents[atomicOperation] = parent;
    }

    public virtual void Reset() {
      Clear();
      parents.Clear();
    }

    public virtual void Generate(IAtomicOperation operation) {
      if (operation == null)
        return;
      Stack<IOperator> trace = new Stack<IOperator>();
      while (operation != null) {
        trace.Push(operation.Operator);
        IAtomicOperation parent = null;
        parents.TryGetValue(operation, out parent);
        operation = parent;
      }
      ReplaceAll(trace);
    }

    #endregion
  }
}
